package com.emsoft.groovy.transform

import com.emsoft.common.IdentityHelper
import groovyjarjarasm.asm.Opcodes
import org.codehaus.groovy.control.CompilePhase
import org.codehaus.groovy.control.SourceUnit
import org.codehaus.groovy.transform.GroovyASTTransformation
import org.codehaus.groovy.ast.*
import org.codehaus.groovy.ast.expr.*
import org.codehaus.groovy.ast.stmt.*

/**
 * Creates equals and hashCode methods.
 *
 * User: Vladimir Zhukov
 */
@GroovyASTTransformation(phase = CompilePhase.CANONICALIZATION)
class IdentityASTTransformation extends BaseASTTransformation {

    private static final CALL_SUPER_MEMBER = "callSuper"
    private static final NULL_EQUALS_MEMBER = "nullEquals"
    private static final IDENTITY_HELPER_TYPE = new ClassNode(IdentityHelper)

    void visit(ASTNode[] nodes, SourceUnit source) {
        init(nodes, source)

        def parent = nodes[1] as AnnotatedNode
        if (parent instanceof FieldNode) {
            def annotation = nodes[0] as AnnotationNode
            def field = parent as FieldNode

            def callSuper = false
            def owner = field.owner
            def ann = annotation.classNode
            def fields = [:]

            owner.fields.each {
                def anns = it.getAnnotations(ann)
                if (anns) {
                    if (RefHelper.getSimpleValue(anns[0], CALL_SUPER_MEMBER) && !callSuper)
                        callSuper = true
                    fields.put(it, anns[0])
                }
            }

            createHashCode(owner, fields, callSuper)
            createEquals(owner, fields, callSuper)
        }
    }

    void createHashCode(ClassNode ownerClass, Map<FieldNode, AnnotationNode> fields, boolean callSuper) {
        if (RefHelper.hasDeclaredMethod(ownerClass, "hashCode"))
            return

        def body = new BlockStatement()

        // _result = IdentityHelper.hashCode(fields)
        def fieldExprs = fields.collect { new VariableExpression(it.key) }
        def args = new TupleExpression(fieldExprs)
        def retSt = new StaticMethodCallExpression(IDENTITY_HELPER_TYPE, "hashCode", args)

        if (callSuper) {
            def result = new VariableExpression("_result")
            body.addStatement(new ExpressionStatement(new DeclarationExpression(result, ASSIGN, retSt)))

            // _result = HashCodeHelper.updateHash(_result, super.hashCode())
            args = new TupleExpression(result, new MethodCallExpression(VariableExpression.SUPER_EXPRESSION,
                    "hashCode", MethodCallExpression.NO_ARGUMENTS))
            def current = new StaticMethodCallExpression(IDENTITY_HELPER_TYPE, "updateHash", args)
            body.addStatement(RefHelper.assignStatement(result, current))
            retSt = result
        }

        body.addStatement(new ReturnStatement(retSt))

        ownerClass.addMethod(new MethodNode("hashCode", Opcodes.ACC_PUBLIC, ClassHelper.int_TYPE,
                Parameter.EMPTY_ARRAY, ClassNode.EMPTY_ARRAY, body))
    }

    void createEquals(ClassNode ownerClass, Map<FieldNode, AnnotationNode> fields, boolean callSuper) {
        if (RefHelper.hasDeclaredMethod(ownerClass, "equals", 1))
            return

        def body = new BlockStatement()
        def other = new VariableExpression("obj")

        // if (obj == null) return false
        body.addStatement(new IfStatement(
                new BooleanExpression(new BinaryExpression(other, COMPARE_EQUAL, ConstantExpression.NULL)),
                new ReturnStatement(ConstantExpression.FALSE), EmptyStatement.INSTANCE)
        )

        // if (this.is(obj)) return true
        body.addStatement(new IfStatement(
                new BooleanExpression(new MethodCallExpression(VariableExpression.THIS_EXPRESSION,
                        "is", new ArgumentListExpression(other))),
                new ReturnStatement(ConstantExpression.TRUE), EmptyStatement.INSTANCE)
        )

        // if (IdentityHelper.equal(this.field, obj.field) && ...) return true
        def equalExprs = fields.keySet().collect {
            def thisField = new VariableExpression(it)
            def args = new TupleExpression(thisField, new PropertyExpression(other, it.getName()),
                    new ConstantExpression(RefHelper.getSimpleValue(fields[it], NULL_EQUALS_MEMBER)))
            new StaticMethodCallExpression(IDENTITY_HELPER_TYPE, "equal", args)
        }

        def binExpr = { list, index ->
            index == list.size() - 1 ? list[index] :
                new BinaryExpression(list[index], LOGICAL_AND, call(list, ++index) as Expression)
        }.call(equalExprs, 0)

        def ifEqSt = new IfStatement(new BooleanExpression(binExpr),
                new ReturnStatement(ConstantExpression.TRUE),
                EmptyStatement.INSTANCE)

        // if (this.getClass() == obj.getClass())
        //     if (IdentityHelper.equal(this.field, obj.field) && ...) return true
        body.addStatement(
                new IfStatement(
                        new BooleanExpression(new BinaryExpression(
                                new MethodCallExpression(VariableExpression.THIS_EXPRESSION, "getClass",
                                        MethodCallExpression.NO_ARGUMENTS),
                                COMPARE_EQUAL,
                                new MethodCallExpression(other, "getClass", MethodCallExpression.NO_ARGUMENTS))),
                        ifEqSt,
                        EmptyStatement.INSTANCE)
        )

        // default
        body.addStatement(new ReturnStatement(ConstantExpression.FALSE))

        ownerClass.addMethod(new MethodNode("equals", Opcodes.ACC_PUBLIC, ClassHelper.boolean_TYPE,
                [new Parameter(ClassHelper.OBJECT_TYPE, "obj")] as Parameter[],
                ClassNode.EMPTY_ARRAY, body))
    }
}
