package com.emsoft.groovy.transform

import com.emsoft.common.Assert
import com.emsoft.groovy.Access
import org.codehaus.groovy.control.CompilePhase
import org.codehaus.groovy.control.SourceUnit
import org.codehaus.groovy.runtime.MetaClassHelper
import org.codehaus.groovy.transform.GroovyASTTransformation
import org.codehaus.groovy.ast.*
import org.codehaus.groovy.ast.expr.*
import org.codehaus.groovy.ast.stmt.*

/**
 * Creates setter if not exist and field not final.
 * Creates getter if not exist.
 * User: Vladimir Zhukov
 *
 * @author Vladimir Zhukov
 *         Date: 21.01.11
 *         Time: 11:13
 */
@GroovyASTTransformation(phase = CompilePhase.CANONICALIZATION)
class PropertyASTTransformation extends BaseASTTransformation {

    private static final def GETTER_MEMBER = "getter"
    private static final def SETTER_MEMBER = "setter"
    private static final def GETTER_FINAL_MEMBER = "getterFinal"
    private static final def SETTER_FINAL_MEMBER = "setterFinal"
    private static final def SETTER_ALLOW_NULL_MEMBER = "setterAllowNull"
    private static final def LAZY_MEMBER = "lazy"

    void visit(ASTNode[] nodes, SourceUnit source) {
        init(nodes, source)

        def parent = nodes[1] as AnnotatedNode
        if (parent instanceof FieldNode) {
            final def annotation = nodes[0] as AnnotationNode
            final def field = parent as FieldNode
            final boolean lazy = ((field.modifiers & ACC_FINAL) == 0 &&
                    (boolean)RefHelper.getSimpleValue(annotation, LAZY_MEMBER) && field.initialValueExpression != null)
            //field.rename('$' + field.name)
            //field.setModifiers(ACC_PRIVATE | (field.getModifiers() & (~(ACC_PUBLIC | ACC_PROTECTED))))
            addGetterIfNeeded(field, annotation, lazy)
            addSetterIfNeeded(field, annotation)
            if (lazy) {
                //@Lazy not meaningful with primitive so convert to wrapper if needed
                if (ClassHelper.isPrimitiveType(field.type))
                    field.setType(ClassHelper.getWrapper(field.type))
            }
        }
    }

    private def getSynchTarget(FieldNode fieldNode) {
        return fieldNode.static ? new ClassExpression(fieldNode.declaringClass) :
            VariableExpression.THIS_EXPRESSION
    }

    private void addGetterIfNeeded(FieldNode fieldNode, AnnotationNode annotation, boolean lazy) {
        final def modifiers = getModifiers(fieldNode, annotation, GETTER_MEMBER, GETTER_FINAL_MEMBER)
        if (modifiers < 0)
            return
        //final def getterName = "get" + MetaClassHelper.capitalize(fieldNode.name.substring(1))
        final def getterName = "get" + MetaClassHelper.capitalize(fieldNode.name)
        if (fieldNode.declaringClass.getGetterMethod(getterName) == null) {
            final def body = new BlockStatement()
            final def fieldExpr = new VariableExpression(fieldNode)
            if (lazy) {
                def assignExpr = new ExpressionStatement(new BinaryExpression(
                        fieldExpr, com.emsoft.groovy.transform.BaseASTTransformation.ASSIGN,
                        fieldNode.initialValueExpression))
                fieldNode.initialValueExpression = null
                def ifTrueStat = (fieldNode.modifiers & ACC_VOLATILE) == 0 ?
                // if not volatile
                    assignExpr :
                // if volatile
                    new SynchronizedStatement(getSynchTarget(fieldNode),
                            new IfStatement(
                                    new BooleanExpression(new BinaryExpression(fieldExpr,
                                            com.emsoft.groovy.transform.BaseASTTransformation.COMPARE_EQUAL,
                                            com.emsoft.groovy.transform.BaseASTTransformation.NULL_EXPR)),
                                    assignExpr,
                                    EmptyStatement.INSTANCE))
                body.addStatement(new IfStatement(
                        new BooleanExpression(new BinaryExpression(fieldExpr,
                                com.emsoft.groovy.transform.BaseASTTransformation.COMPARE_EQUAL,
                                com.emsoft.groovy.transform.BaseASTTransformation.NULL_EXPR)),
                        ifTrueStat,
                        EmptyStatement.INSTANCE))
            }
            body.addStatement(new ReturnStatement(fieldExpr))
            fieldNode.declaringClass.addMethod(getterName, modifiers, fieldNode.type,
                    Parameter.EMPTY_ARRAY, ClassNode.EMPTY_ARRAY, body)
        }
    }

    private void addSetterIfNeeded(FieldNode fieldNode, AnnotationNode annotation) {
        final def modifiers = getModifiers(fieldNode, annotation, SETTER_MEMBER, SETTER_FINAL_MEMBER)
        if (modifiers < 0)
            modifiers = ACC_PRIVATE // because groovy generate setter if not present
        //final def setterName = "set" + MetaClassHelper.capitalize(fieldNode.name.substring(1))
        final def setterName = "set" + MetaClassHelper.capitalize(fieldNode.name)
        if ((fieldNode.modifiers & ACC_FINAL) == 0 &&
                fieldNode.declaringClass.getSetterMethod(setterName) == null) {
            // for allow set null in setter
            def allowNull = RefHelper.getSimpleValue(annotation, SETTER_ALLOW_NULL_MEMBER).equals(true)
            // parameter name
            final def varName = "value"
            // method body
            final def body = new BlockStatement()
            if (!allowNull)
                body.addStatement(new ExpressionStatement(new MethodCallExpression(
                        new ClassExpression(new ClassNode(Assert)), "notNull",
                        new ArgumentListExpression(
                                new VariableExpression(varName),
                                new ConstantExpression(varName)))))
            body.addStatement(new ExpressionStatement(new BinaryExpression(
                    new FieldExpression(fieldNode), com.emsoft.groovy.transform.BaseASTTransformation.ASSIGN,
                    new VariableExpression(varName))))
            fieldNode.declaringClass.addMethod(setterName, modifiers, ClassHelper.VOID_TYPE,
                    [new Parameter(fieldNode.type, varName)] as Parameter[], ClassNode.EMPTY_ARRAY, body)
        }
    }

    private int getModifiers(FieldNode field, AnnotationNode annotation, String accessMember, String finalMember) {
        def modifiers = 0
        if (field.isStatic())
            modifiers |= ACC_STATIC
        if (RefHelper.getSimpleValue(annotation, finalMember).equals(true))
            modifiers |= ACC_FINAL

        def accessExp = RefHelper.getPropertyValue(annotation, accessMember)
        def access = Enum.valueOf(Access.class, (accessExp instanceof PropertyExpression ?
            accessExp.propertyAsString : accessExp as String))

        switch (access) {
            case Access.PACKAGE:
                modifiers |= 0
                break
            case Access.PROTECTED:
                modifiers |= ACC_PROTECTED
                break
            case Access.PRIVATE:
                modifiers |= ACC_PRIVATE
                break
            case Access.PUBLIC:
                modifiers |= ACC_PUBLIC
                break
            default:
                modifiers = -1
                break
        }

        return modifiers
    }
}
